# -*- coding: utf-8 -*-
import datetime
import logging
import tempfile
import zipfile
from zipfile import ZIP_DEFLATED
from django.conf import settings
from django.contrib.auth.decorators import login_required
from django.db import connection
from django.http import HttpResponse
from django.shortcuts import render
from django.views.decorators.http import require_POST
from django.utils.datastructures import MultiValueDictKeyError
from openpyxl import load_workbook
from dates2utils.dates2export import (
    xml_teilnehmerDatenImport,
    update_by_timeinterval
)
from reportserver.lib.io.requests import (
    check_date,
    convert_german_date_to_iso
)


logger = logging.getLogger(__name__)


def dict_cursor(cursor):
    description = [x[0] for x in cursor.description]
    for row in cursor:
        yield dict(zip(description, row))


def extract_program_name(beneficiary_id):
    """Extract the program name from beneficiary_id.

    By definition the first section, seperated by '.' (dot).
    """
    return beneficiary_id.split('.')[0]


def update_indicator_name(row, indicators):
    """It is unclear why the indicator name update is necessary.

    This should be explained.
    """
    r = {}
    for key in row:
        if key.upper() in indicators:
            r[key.upper()] = row[key]
        elif key.upper().replace("_", ".") in indicators:
            r[key.upper().replace("_", ".")] = row[key]
        else:
            r[key] = row[key]
    return r


def personal_id_to_instanceid(value, separator="-", merger="."):
    split_value = value.split(separator)
    return merger.join(split_value[0:2])


def stringify_pair(p1, p2):
    return "%s;%s" % (p1, p2)


def get_full_beneficiary(string1, beneficiaries):
    # calc the year number of the beneficiary_id
    for beneficiary in beneficiaries:
        instance_id_with_underscores = beneficiary[0]
        if instance_id_with_underscores.startswith(string1):
            return instance_id_with_underscores
    return string1


def get_xml_file_name(orig_bene_id, time_str, rc=None):
    bene_id_with_underscores = orig_bene_id.replace(".", "_")
    if rc:
        return u"{}_{}_{}.xml".format(bene_id_with_underscores, rc, time_str)
    else:
        return u"{}_{}.xml".format(bene_id_with_underscores, time_str)


def get_GUID(orig_bene_id, session, rc="*"):
    # Crash if no entry in LUT, i.e. beneficiary referenced in
    # participations but not considered for DATES II export.
    guid = session.get("dates2_lut")[stringify_pair(orig_bene_id, rc)]

    # Process marker, if no guid for known beneficiary defined
    if guid == "NOCH-KEINE-GUID-FESTGELEGT":
        return None
    else:
        return guid


def filter_only_rc(rows, rc):
    return [row for row in rows if row.get('regioncategory_id') == rc]


def get_virtual_participations(beneficiary_id, cursor, start, end):
    cursor.execute("SELECT * FROM eval_virtual_participations"
                   " WHERE originating_beneficiary_id=%s ;",
                   [beneficiary_id])
    rows = [x for x in dict_cursor(cursor)]
    rows = update_by_timeinterval(
        rows, start, end,
        {'RI': settings.DATES2_RI_INDICATORS})
    return rows


def handle_beneficiary(beneficiary_id, cursor, start, end, time_str, zip1,
                       session):
    rows = get_virtual_participations(beneficiary_id, cursor, start, end)
    out = []
    oi_indicators = settings.DATES2_OI_INDICATORS
    ri_indicators = settings.DATES2_RI_INDICATORS
    indicators = oi_indicators + ri_indicators
    rows_updated = [update_indicator_name(row, indicators)
                    for row in rows]
    if settings.PROGRAM == "EC":
        for rc in [1, 2, 3, 4]:
            out = []
            rc_filtered_rows = filter_only_rc(rows_updated, rc)
            if not rc_filtered_rows:
                continue
            xml_teilnehmerDatenImport(out,
                                      get_GUID(beneficiary_id, session, rc),
                                      rc_filtered_rows,
                                      oi_indicators, ri_indicators)
            output = u"\n".join(out)
            zip1.writestr(get_xml_file_name(beneficiary_id, time_str, rc),
                          output)
    else:
        out = []
        xml_teilnehmerDatenImport(out, get_GUID(beneficiary_id, session),
                                  rows_updated,
                                  oi_indicators, ri_indicators)
        output = u"\n".join(out)
        zip1.writestr(get_xml_file_name(beneficiary_id, time_str), output)


def get_date_from(request, name):
    return datetime.datetime.strptime(check_date(convert_german_date_to_iso(
        request.POST.get(name))), '%Y-%m-%d').date()


def iso_date_to_german_date(date_str):
    date1 = datetime.datetime.strptime(date_str, '%Y-%m-%d')
    return date1.strftime('%d.%m.%Y')


def handle_insert_lists(cursor, start, end, time_str, session, main_zip):
    # handle dates2 insert xml part
    cursor.execute("SELECT DISTINCT originating_beneficiary_id "
                   "FROM eval_virtual_participations;")
    beneficiary_ids = [row[0] for row in cursor.fetchall()]
    with tempfile.TemporaryFile() as ziptmpfile2:
        with zipfile.ZipFile(ziptmpfile2, "w", ZIP_DEFLATED) as zip2:
            # for every b: 1. get rows 2. xml them
            for beneficiary_id in beneficiary_ids:
                # Process beneficiaries with existing GUID
                # This is the case for all in EC so pass the test.
                if settings.PROGRAM == "EC" \
                        or get_GUID(beneficiary_id, session):
                    handle_beneficiary(beneficiary_id, cursor, start, end,
                                       time_str, zip2, session)
        ziptmpfile2.seek(0)
        progname = get_progname_by_beneficiaries(beneficiary_ids)
        main_zip.writestr(create_normal_zip_arcname(progname, time_str),
                          ziptmpfile2.read())


def create_normal_zip_arcname(program, time_str):
    """create the arcname for the normal xml zip."""
    return "Teilnehmerlisten-%s_%s.zip" % (program, time_str)


# code for generation of the remove lists
def create_rm_list_xml_arcname(beneficiary, time_str, two_ways_differ):
    """create the arcname for a rm list xml."""
    format_str = u"%s_%s.xml"
    if two_ways_differ:
        format_str = u"%s_LL-reduziert_%s.xml"
    return format_str % (beneficiary, time_str)


def create_rm_lists_zip_arcname(program, time_str, region_category,
                                based_only_way1):
    """create the arcname for a rm lists zip."""
    if region_category == "*":
        format_str = "Loeschlisten-%s_%s.zip"
        if based_only_way1:
            format_str = "Loeschlisten-%s_LL-reduziert_%s.zip"
        return format_str % (program, time_str)
    format_str = "Loeschlisten-%s-%s_%s.zip"
    if based_only_way1:
        format_str = "Loeschlisten-%s-%s_LL-reduziert_%s.zip"
    return format_str % (program, region_category, time_str)


def generate_rm_list_xml(pre_rm_list, instance_guid):
    out = []
    xml_teilnehmerDatenImport(out, instance_guid, [], [], [],
                              loeschliste=pre_rm_list)
    output = u"\n".join(out)
    return output


def handle_rm_lists(cursor, start, end, time_str, main_zip, session,
                    existing_pts):
    """Handle the create of the rm lists (XML-Löschlisten), write the Zips
    and insert them into the main_zip. Assume less 2000 instances and 2000
    participants per instance."""
    cursor.execute("SELECT uuid, personal_id, originating_beneficiary_id"
                   " FROM participant_ids ;")
    rows = [row for row in cursor.fetchall()]
    orig_bene_ids = set([row[2] for row in rows])
    if settings.PROGRAM == "EC":
        region_categories = ["1", "2", "3", "4"]
    else:
        region_categories = ["*"]

    # check if one of the rmlists is only based on way 1
    based_only_way1 = False

    for region_category in region_categories:
        with tempfile.TemporaryFile() as ziptmpfile:
            ZDE = ZIP_DEFLATED
            with zipfile.ZipFile(ziptmpfile, "w", ZDE) as secondary_zip:

                # for all orig_bene ids in the rm lists, create a rmlist.
                for orig_bene_id in orig_bene_ids:

                    two_ways_differ = False

                    # get instance_guid and instance_id_fulll
                    if settings.PROGRAM == "EC":
                        instance_guid = get_GUID(orig_bene_id, session,
                                                 region_category)
                        instance_id_full = (orig_bene_id + "-" +
                                            region_category)
                    else:
                        instance_guid = get_GUID(orig_bene_id, session)
                        instance_id_full = orig_bene_id

                    if not instance_guid:
                        # Skip beneficiary, if no GUID set.
                        continue

                    # Generate pre rm list in two ways and check

                    # helper
                    def create_entry(uuid, pid):
                        return {"participant_uuid": uuid, "personal_id": pid}

                    def check_both_ways(pre_rm_list1, pre_rm_list2):
                        pre_rm_set1 = set(pre_rm_list1)
                        pre_rm_set2 = set(pre_rm_list2)
                        if pre_rm_set1 != pre_rm_set2:
                            logger.error("Pre rm lists differ.")
                            logger.error(pre_rm_list1)
                            logger.error(pre_rm_list2)
                            if not settings.DATES2_SUPPRESS_RMLIST_ERROR:
                                raise ValueError(u"Bei Vorhaben '{}' gibt es"
                                                 u" Probleme beim Berechnen "
                                                 u"der Löschlisten.".format(
                                                     orig_bene_id))
                            return False
                        return True

                    # 1. way of generating pre_rm_list
                    pre_rm_list1 = [(row[0], row[1])
                                    for row in rows
                                    if row[2] == orig_bene_id and
                                    (row[0], row[1], row[0]) in
                                    existing_pts.get()]

                    # 2.way of generating pre_rm_list
                    virtual_participations = get_virtual_participations(
                        orig_bene_id, cursor, start, end)
                    vpns = set([(vpn['participant_uuid'],
                                 vpn['personal_id'],
                                 vpn['participant_uuid'])
                                for vpn in virtual_participations])
                    pre_rm_list2 = [(row[0], row[1])
                                    for row in
                                    existing_pts.get_by_instance_guid(
                                        instance_guid)
                                    if row not in vpns]

                    # check if equal.
                    if not check_both_ways(pre_rm_list1, pre_rm_list2):
                        based_only_way1 = True
                        two_ways_differ = True

                    # generate xml
                    rm_list = [create_entry(item[0], item[1]) for item in
                               pre_rm_list1]
                    xml = generate_rm_list_xml(rm_list, instance_guid)

                    # write xml into the archive
                    arcname = create_rm_list_xml_arcname(
                        instance_id_full, time_str, two_ways_differ)
                    secondary_zip.writestr(arcname, xml)
            ziptmpfile.seek(0)
            zip_str = ziptmpfile.read()

            # write zip result into main zip as Löschlisten-*.zip
            progname = get_progname_by_beneficiaries(
                [bene[0] for bene in session['beneficiary']])
            zip_arcname = create_rm_lists_zip_arcname(progname,
                                                      time_str,
                                                      region_category,
                                                      based_only_way1)
            main_zip.writestr(zip_arcname, zip_str)


class ParticipationsInDates2(object):
    """includes the participations from dates2"""

    def __init__(self):
        self._list = []

    def append(self, participation_tuple):
        """Tuple should be (instance_guid, participant_uuid, personal_id,
                            participant_uuid)."""
        self._list.append(participation_tuple)

    def get(self):
        """Get all pns. Return list of (uuid, personal_id, uuid) tripple."""
        return [(val[1], val[2], val[3]) for val in self._list]

    def get_by_instance_guid(self, instance_id):
        """Return list (uuid, personal_id, uuid) if the right instance_id."""
        return [(val[1], val[2], val[3]) for val in self._list
                if val[0] == instance_id]


def read_from_xlsx_file(pn_overview_file):
    workbook = load_workbook(filename=pn_overview_file)
    sheet = workbook.active
    result = ParticipationsInDates2()
    for nr, row in enumerate(sheet.rows):
        if nr < 5:
            continue
        if row[0].value:
            result.append((row[0].value, row[1].value, row[2].value,
                           row[3].value))
    return result


# end rm list stuff

def get_progname_by_beneficiaries(beneficiary_ids):
    if beneficiary_ids:
        return extract_program_name(beneficiary_ids[0])
    else:
        return settings.PROGRAM


@login_required
@require_POST
def export_dates2(request):
    time_str = datetime.datetime.now().strftime("%Y%m%d_%H%M%S")
    try:
        start = get_date_from(request, 'start_date')
        end = get_date_from(request, 'end_date')
        pn_overview_file = request.FILES['pn_overview_file']
        no_pn_overview_file = False
    except ValueError:
        return overview(request)
    except MultiValueDictKeyError:
        no_pn_overview_file = True

    if not no_pn_overview_file:
        try:
            existing_pts = read_from_xlsx_file(pn_overview_file)
        except Exception as e:
            logger.exception(e)
            return overview(request, error_msg=(u"Beim Lesen der TN-Übersicht"
                                                u" ist"
                                                u" ein Fehler aufgetreten."),
                            end_date=end.strftime('%d.%m.%Y'))

    with connection.cursor() as cursor:
        with tempfile.TemporaryFile() as ziptmpfile:
            with zipfile.ZipFile(ziptmpfile, "w", ZIP_DEFLATED) as zip1:

                if not no_pn_overview_file:
                    try:
                        handle_rm_lists(cursor, start, end, time_str, zip1,
                                        request.session, existing_pts)
                    except ValueError as vale:
                        return overview(request, error_msg=vale.message,
                                        end_date=end.strftime('%d.%m.%Y'))
                handle_insert_lists(cursor, start, end, time_str,
                                    request.session, zip1)
            ziptmpfile.seek(0)
            zip_str = ziptmpfile.read()

    program_name = get_progname_by_beneficiaries(
        [bene[0] for bene in request.session.get('beneficiary')])
    response = HttpResponse(zip_str, content_type="application/zip")
    response['Content-Disposition'] = ('attachment; filename="%s_%s.zip"' %
                                       (program_name, time_str))
    logger.info("'%s' requested a dates2export.", request.user.username)
    return response


@login_required
def overview(request, error_msg=None, end_date=None):
    default_end_date = iso_date_to_german_date(
        settings.DATES2_DEFAULT_END_DATE)
    if end_date:
        default_end_date = end_date
    return render(request, 'dates2export.html',
                  {"default_end_date": default_end_date,
                   "error_msg": error_msg})
